"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const pattern_1 = require("./pattern");
const XRegExp = require("xregexp");
const regexpUtil_1 = require("./regexpUtil");
const cacheMap_1 = require("../util/cacheMap");
const maxDepth = 100;
const useLogging = false;
function logInfo(message) {
    useLogging && console.log(message);
}
function tokenizeLine(text, rule) {
    const tokens = [];
    let offset = 0;
    let end = rule.end;
    let endMatch = end ? XRegExp.exec(text, end, offset) : undefined;
    let endOffset = endMatch ? endMatch.index : text.length;
    while (offset < text.length) {
        const { match, rule: matchingRule } = matchRule(text, offset, rule);
        if (match && match.index <= endOffset) {
            if (match.index > offset) {
                tokens.push({ startIndex: offset, endIndex: match.index, scopes: extractScopes(rule) });
            }
            tokens.push(...tokenizeCapture(matchingRule, match, pattern_1.captures(matchingRule.pattern)));
            offset = match.index + match[0].length;
            const pattern = matchingRule.pattern;
            if (pattern_1.isPatternBeginEnd(pattern)) {
                rule = {
                    parent: matchingRule,
                    pattern: { patterns: pattern.patterns || [] },
                    grammarDef: matchingRule.grammarDef,
                    depth: matchingRule.depth + 1,
                    scope: pattern.contentName,
                    end: buildEndRegEx(pattern.end, match),
                    comment: `Begin ${rule.depth + 1}: ${pattern.begin} <--> ${pattern.end} # ` + (pattern.name || pattern.comment || ''),
                };
                end = rule.end;
                try {
                    endMatch = end ? XRegExp.exec(text, end, offset) : undefined;
                    endOffset = endMatch ? endMatch.index : text.length;
                }
                catch (e) {
                    console.log(e);
                    endMatch = undefined;
                    endOffset = text.length;
                }
                continue;
            }
        }
        else {
            if (offset < endOffset) {
                tokens.push({ startIndex: offset, endIndex: endOffset, scopes: extractScopes(rule) });
            }
            offset = endOffset;
        }
        // We need to pop out and process any applicable ending patterns.
        do {
            rule = findEndRule(rule);
            end = rule.end;
            try {
                endMatch = end ? XRegExp.exec(text, end, offset) : undefined;
                endOffset = endMatch ? endMatch.index : text.length;
            }
            catch (e) {
                console.log(e);
                endMatch = undefined;
                endOffset = text.length;
            }
            if (offset === endOffset) {
                // process ending rule.
                if (rule.parent && endMatch) {
                    rule = findBoundingRule(rule);
                    tokens.push(...tokenizeCapture(rule, endMatch, pattern_1.endCaptures(rule.pattern)));
                    offset = endMatch.index + endMatch[0].length;
                }
            }
        } while (rule.parent && endMatch && offset >= endOffset);
    }
    return {
        tokens,
        state: rule,
    };
}
exports.tokenizeLine = tokenizeLine;
function matchRule(text, offset, rule) {
    let result;
    try {
        logInfo(`${'.'.repeat(rule.depth)}+${rule.depth} ${pattern_1.patternToString(rule.pattern)} test`);
        result = matchRuleInner(text, offset, rule);
        return result;
    }
    finally {
        const msg = result
            ? (result.match ? `match ${result.match.length}` : 'non-match')
            : 'failed';
        logInfo(`${'.'.repeat(rule.depth)}+${rule.depth} ${pattern_1.patternToString(rule.pattern)} result ${msg}`);
    }
}
exports.matchRule = matchRule;
function matchRuleInner(text, offset, rule) {
    const { pattern, depth, grammarDef } = rule;
    if (pattern_1.isPatternInclude(pattern)) {
        if (depth < maxDepth && grammarDef.repository) {
            const name = pattern.include.slice(1);
            const result = pattern._reference || grammarDef.repository[name];
            if (result) {
                return matchRule(text, offset, {
                    grammarDef,
                    pattern: result,
                    parent: rule,
                    depth: depth + 1,
                    scope: pattern_1.scope(result),
                    comment: 'Include: ' + (pattern.comment || name),
                });
            }
            if (pattern.include === '$self') {
                return matchRule(text, offset, grammarToRule(grammarDef, rule));
            }
            // We do not support including other grammars yet.
            console.log(`Unknown include: (${name})`);
        }
        return { rule }; // Unsupported include, match nothing.
    }
    if (pattern_1.isPatternMatch(pattern)) {
        const { regex, sticky } = regExpOrStringToRegExp(pattern.match);
        try {
            return { match: XRegExp.exec(text, regex, offset, sticky), rule };
        }
        catch (e) {
            console.log(e);
            return { rule };
        }
    }
    if (pattern_1.isPatternBeginEnd(pattern)) {
        try {
            const { regex, sticky } = regExpOrStringToRegExp(pattern.begin);
            return { match: XRegExp.exec(text, regex, offset, sticky), rule };
        }
        catch (e) {
            console.log(e);
            return { rule };
        }
    }
    if (pattern_1.isPatternName(pattern)) {
        return { match: XRegExp.exec(text, /.*/, offset, false), rule };
    }
    let best;
    for (const pat of pattern.patterns) {
        const m = matchRule(text, offset, {
            grammarDef,
            pattern: pat,
            parent: rule,
            depth: depth + 1,
            scope: pattern_1.scope(pat),
            comment: `Nested ${rule.comment}: ` + (pat.name || pat.comment || '*'),
        });
        if (!best) {
            best = m;
        }
        if (m.match) {
            if (m.match.index === offset) {
                return m;
            }
            if (!best.match || m.match.index < best.match.index) {
                best = m;
            }
        }
    }
    if (!best) {
        best = {
            match: undefined,
            rule
        };
    }
    return best;
}
function findBoundingRule(rule) {
    while (rule.parent && !pattern_1.isPatternBeginEnd(rule.pattern)) {
        rule = rule.parent;
    }
    return rule;
}
function findEndRule(rule) {
    while (rule.parent && !rule.end) {
        rule = rule.parent;
    }
    return rule;
}
const matchSlashG = /^\\G/;
const matchNegSlashG = '(?!\\G)';
const regExCache = cacheMap_1.create(_regExpOrStringToRegExp);
function regExpOrStringToRegExp(regex) {
    return regExCache.get(regex);
}
function _regExpOrStringToRegExp(regex) {
    try {
        if (typeof regex === 'string') {
            if (matchSlashG.test(regex)) {
                return { regex: XRegExp(regex.replace(matchSlashG, '')), sticky: true };
            }
            return { regex: XRegExp(regex) };
        }
        return { regex: XRegExp(regex) };
    }
    catch (e) {
        console.log(e);
        return { regex: /.^/ };
    }
}
function buildEndRegEx(regex, match) {
    if (!match) {
        return undefined;
    }
    if (typeof regex === 'string') {
        const subs = regexpUtil_1.escapeMatch(match);
        try {
            if (regex.startsWith(matchNegSlashG)) {
                regex = '(?<=.|^)|$';
            }
            return XRegExp.build(regex, subs);
        }
        catch (e) {
            console.log(e);
            return /.^/;
        }
    }
    return regex;
}
function extractScopes(rule) {
    const values = [];
    let r = rule;
    while (r) {
        const { name = undefined } = r.pattern;
        const scope = r.scope || name || undefined;
        if (scope) {
            values.push(...scope.split(' ').reverse());
        }
        r = r.parent;
    }
    return values.reverse();
}
function tokenizeCapture(rule, match, cap) {
    const scopes = extractScopes(rule);
    let startIndex = match.index;
    const endIndex = startIndex + match[0].length;
    const tokens = [{ startIndex, endIndex, scopes }];
    if (cap) {
        const text = match.input;
        const captures = pairCaptureGroupsToMatchOffsets([...Object.keys(cap)], regexpUtil_1.matchesToOffsets(match));
        const capturedTokens = captures.map(g => {
            const patterns = g.captureGroups.filter(v => !!cap[v]).map(v => cap[v]);
            const capturedScopes = patterns
                .filter(pattern_1.isPatternName)
                .map(p => p.name.split(' '))
                .reduce((a, b) => a.concat(b), scopes);
            if (!capturedScopes.length) {
                return [];
            }
            const pattern = patterns[patterns.length - 1];
            if (pattern_1.isPatternName(pattern)) {
                return [{
                        startIndex: g.begin,
                        endIndex: g.end,
                        scopes: capturedScopes,
                    }];
            }
            return tokenizeLine(text.slice(g.begin, g.end), {
                parent: undefined,
                pattern,
                grammarDef: rule.grammarDef,
                depth: 0,
                scope: capturedScopes.join(' '),
                comment: 'Capture: ' + (pattern.name || pattern.comment),
            }).tokens.map(t => (Object.assign({}, t, { startIndex: t.startIndex + g.begin, endIndex: t.endIndex + g.begin })));
        }).reduce((a, b) => a.concat(b), []);
        tokens.push(...capturedTokens);
    }
    return mergeTokens(tokens);
}
// mergeTokens will merge overlapping tokens.
// tokens later in the list always win.
function mergeTokens(tokens) {
    const result = [];
    function findStart(startIndex) {
        // Search backwards in the result list.
        let p;
        for (p = result.length; p > 0 && result[p - 1].endIndex > startIndex; --p) { }
        return p;
    }
    function findEnd(endIndex, fromPos) {
        let p;
        for (p = fromPos; p < result.length && result[p].startIndex < endIndex; ++p) { }
        return p;
    }
    for (const t of tokens.filter(t => t.endIndex > t.startIndex)) {
        const pos = findStart(t.startIndex);
        if (pos === result.length) {
            // Just append
            result.push(t);
            continue;
        }
        const endPos = findEnd(t.endIndex, pos);
        if (endPos === pos) {
            result.splice(pos, 0, t);
            continue;
        }
        const overlappingTokens = result.slice(pos, endPos);
        const newTokens = [];
        const front = overlappingTokens[0];
        const back = overlappingTokens[overlappingTokens.length - 1];
        if (front.startIndex < t.startIndex) {
            newTokens.push(Object.assign({}, front, { endIndex: t.startIndex }));
        }
        newTokens.push(t);
        if (back.endIndex > t.endIndex) {
            newTokens.push(Object.assign({}, back, { startIndex: t.endIndex }));
        }
        result.splice(pos, endPos - pos, ...newTokens);
    }
    return result;
}
exports.mergeTokens = mergeTokens;
function pairCaptureGroupsToMatchOffsets(captureGroups, match) {
    const filteredGroups = captureGroups
        .filter(g => match.has(g))
        .map(captureGroup => (Object.assign({}, match.get(captureGroup), { captureGroup })));
    const foundMatches = filteredGroups.map(g => {
        return Object.assign({}, g, { captureGroups: filteredGroups.filter(sg => sg.end >= g.end && sg.begin <= g.begin).map(sg => sg.captureGroup) });
    });
    return foundMatches;
}
exports.pairCaptureGroupsToMatchOffsets = pairCaptureGroupsToMatchOffsets;
function grammarToRule(grammar, parent) {
    if (parent) {
        return {
            parent,
            pattern: { patterns: grammar.patterns },
            scope: undefined,
            grammarDef: grammar,
            depth: parent.depth + 1,
            comment: 'Nested Grammar' + ' ' + (grammar.name || grammar.scopeName || ''),
        };
    }
    const rule = {
        parent,
        pattern: grammar,
        scope: grammar.scopeName,
        grammarDef: grammar,
        depth: 0,
        comment: 'Root ' + (grammar.name || grammar.scopeName || ''),
    };
    return rule;
}
exports.grammarToRule = grammarToRule;
//# sourceMappingURL=tokenize.js.map